# define parser
# written by MOROHOSHI Akihiko <moro@nii.ac.jp>
# $Id: smlngparser.rb,v 1.3 2001/07/28 20:49:49 moro Exp $

require 'context'

########################################################################
# constants
SPC = ' '
SPACE = '[ \r\n\t]+'

########################################################################
# Token
class Token
    def initialize(type, text, context)
	@type, @text, @context = type, text, context
	@context.freeze
    end
    attr_reader :type, :text, :context
    def space_context
	@space_context ||= @context.space_context.freeze
    end

    def append(text)
	@text += text
	unless @context.typewriter
	    @text.gsub! /#{SPACE}/,SPC
	end
    end

    ### property entries used by optimizer.
    attr_accessor :estimated_remaining_cost
end

########################################################################
# optimizer class
class SMLNGParser
    def initialize(text)
	@scanner = SMLNGScanner.new(text)
	@context_stack = [ [nil, Context.new] ]
	@current_context = @context_stack[0][1]
	@tokens = []
	@last_space = nil
    end
    attr_reader :tokens

    def tag_length
	@scanner.tag_length
    end

    def parse
	while a = @scanner.scan
	    sym,text = a
	    case sym
	    when :OPENTAG
		open_tag(text)
	    when :CLOSETAG
		close_tag(text)
	    when :SPACE
		add_space(text)
	    when :TEXT
		add_text(text)
	    else
		raise "unknown sym: #{sym}"
	    end
	end
	if @context_stack.size > 1
	    raise "malformed document; close tag expected.\n"
	end
	@tokens
    end

    ####################################################################
    #
    private
    
    def open_tag(name)
	cont = context_modified_by_tag(@current_context, name)
	@context_stack.push [name,cont]
	@current_context = cont
    end

    def close_tag(text)
	tag,cont = @context_stack.pop
	if tag.nil?
	    raise "unexpected close tag: </#{text}>"
	elsif tag != text
	    raise "close tag mismatch: <#{tag}> -- </#{text}>"
	end
	@current_context = @context_stack[-1][1]
    end

    def add_space(text)
	context = @current_context
	unless context.typewriter
	    text = SPC
	end
	token = Token.new(:SPACE, text, context)
	if @last_space
	    if @last_space.space_context == token.space_context
		if token.space_context.typewriter
		    @last_space.append text
		    return
		else
		    return	# whitespace simplification
		end
	    end
	elsif @last_token
	    if @last_token.type == :TEXT and
		    @last_token.space_context == context.space_context
		@last_token.append text
		return
	    end
	end

	add_token token
    end

    def add_text(text)
	context = @current_context
	unless context.typewriter
	    text.gsub! /#{SPACE}/o, SPC
	end
	if @last_token
	    if @last_token.type == :SPACE
		if @last_token.space_context == context.space_context
		    replace_last_token Token.new(:TEXT,
						 @last_token.text + text,
						 context)
		    return
		end
	    else		# :TEXT
		if @last_token.context == context
		    @last_token.append text
		    return
		end
	    end
	end
	add_token Token.new(:TEXT, text, context)
    end

    def add_token(token)
	@tokens << token
	@last_token = token
	if token.type == :SPACE or /#{SPACE}\Z/o =~ token.text
	    @last_space = token
	else
	    @last_space = nil
	end
    end

    def replace_last_token(token)
	@tokens.pop
	add_token(token)
    end
end
